"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const graphql_1 = require("graphql");
const graphql = __importStar(require("graphql"));
const graphql_extensions_1 = require("graphql-extensions");
const apollo_cache_control_1 = require("apollo-cache-control");
const apollo_tracing_1 = require("apollo-tracing");
const apollo_server_errors_1 = require("apollo-server-errors");
const crypto_1 = require("crypto");
const requestPipelineAPI_1 = require("./requestPipelineAPI");
exports.InvalidGraphQLRequestError = requestPipelineAPI_1.InvalidGraphQLRequestError;
class GraphQLRequestPipeline {
    constructor(config) {
        this.config = config;
        graphql_extensions_1.enableGraphQLExtensions(config.schema);
    }
    processRequest(requestContext) {
        return __awaiter(this, void 0, void 0, function* () {
            const config = this.config;
            const requestListeners = [];
            if (config.plugins) {
                for (const plugin of config.plugins) {
                    if (!plugin.requestDidStart)
                        continue;
                    const listener = plugin.requestDidStart(requestContext);
                    if (listener) {
                        requestListeners.push(listener);
                    }
                }
            }
            const extensionStack = this.initializeExtensionStack();
            requestContext.context._extensionStack = extensionStack;
            this.initializeDataSources(requestContext);
            yield Promise.all(requestListeners.map(listener => listener.prepareRequest && listener.prepareRequest(requestContext)));
            const request = requestContext.request;
            let { query, extensions } = request;
            let persistedQueryHit = false;
            let persistedQueryRegister = false;
            if (!query) {
                if (extensions && extensions.persistedQuery) {
                    if (!this.config.persistedQueries ||
                        !this.config.persistedQueries.cache) {
                        throw new apollo_server_errors_1.PersistedQueryNotSupportedError();
                    }
                    else if (extensions.persistedQuery.version !== 1) {
                        throw new requestPipelineAPI_1.InvalidGraphQLRequestError('Unsupported persisted query version');
                    }
                    const sha = extensions.persistedQuery.sha256Hash;
                    if (query === undefined) {
                        query =
                            (yield this.config.persistedQueries.cache.get(`apq:${sha}`)) ||
                                undefined;
                        if (query) {
                            persistedQueryHit = true;
                        }
                        else {
                            throw new apollo_server_errors_1.PersistedQueryNotFoundError();
                        }
                    }
                    else {
                        const hash = crypto_1.createHash('sha256');
                        const calculatedSha = hash.update(query).digest('hex');
                        if (sha !== calculatedSha) {
                            throw new requestPipelineAPI_1.InvalidGraphQLRequestError('provided sha does not match query');
                        }
                        persistedQueryRegister = true;
                        (() => __awaiter(this, void 0, void 0, function* () {
                            return (this.config.persistedQueries &&
                                this.config.persistedQueries.cache.set(`apq:${sha}`, query));
                        }))().catch(error => {
                            console.warn(error);
                        });
                    }
                }
            }
            if (!query) {
                throw new requestPipelineAPI_1.InvalidGraphQLRequestError('Must provide query string.');
            }
            const requestDidEnd = extensionStack.requestDidStart({
                request: request.http,
                queryString: request.query,
                operationName: request.operationName,
                variables: request.variables,
                extensions: request.extensions,
                persistedQueryHit,
                persistedQueryRegister,
            });
            try {
                let document;
                try {
                    document = parse(query);
                }
                catch (syntaxError) {
                    return sendResponse({
                        errors: [
                            apollo_server_errors_1.fromGraphQLError(syntaxError, {
                                errorClass: apollo_server_errors_1.SyntaxError,
                            }),
                        ],
                    });
                }
                const validationErrors = validate(document);
                if (validationErrors.length > 0) {
                    return sendResponse({
                        errors: validationErrors.map(validationError => apollo_server_errors_1.fromGraphQLError(validationError, {
                            errorClass: apollo_server_errors_1.ValidationError,
                        })),
                    });
                }
                const operation = graphql_1.getOperationAST(document, request.operationName);
                if (operation && this.willExecuteOperation) {
                    this.willExecuteOperation(operation);
                }
                requestContext.operation = operation;
                requestListeners.forEach(listener => listener.executionDidStart &&
                    listener.executionDidStart(requestContext));
                let response;
                try {
                    response = (yield execute(document, request.operationName, request.variables));
                }
                catch (executionError) {
                    return sendResponse({
                        errors: [apollo_server_errors_1.fromGraphQLError(executionError)],
                    });
                }
                const formattedExtensions = extensionStack.format();
                if (Object.keys(formattedExtensions).length > 0) {
                    response.extensions = formattedExtensions;
                }
                if (this.config.formatResponse) {
                    response = this.config.formatResponse(response, {
                        context: requestContext.context,
                    });
                }
                return sendResponse(response);
            }
            finally {
                requestDidEnd();
            }
            function parse(query) {
                const parsingDidEnd = extensionStack.parsingDidStart({
                    queryString: query,
                });
                try {
                    return graphql.parse(query);
                }
                finally {
                    parsingDidEnd();
                }
            }
            function validate(document) {
                let rules = graphql_1.specifiedRules;
                if (config.validationRules) {
                    rules = rules.concat(config.validationRules);
                }
                const validationDidEnd = extensionStack.validationDidStart();
                try {
                    return graphql.validate(config.schema, document, rules);
                }
                finally {
                    validationDidEnd();
                }
            }
            function execute(document, operationName, variables) {
                return __awaiter(this, void 0, void 0, function* () {
                    const executionArgs = {
                        schema: config.schema,
                        document,
                        rootValue: typeof config.rootValue === 'function'
                            ? config.rootValue(document)
                            : config.rootValue,
                        contextValue: requestContext.context,
                        variableValues: variables,
                        operationName,
                        fieldResolver: config.fieldResolver,
                    };
                    const executionDidEnd = extensionStack.executionDidStart({
                        executionArgs,
                    });
                    try {
                        return graphql.execute(executionArgs);
                    }
                    finally {
                        executionDidEnd();
                    }
                });
            }
            function sendResponse(response) {
                return (requestContext.response = extensionStack.willSendResponse({
                    graphqlResponse: Object.assign({}, requestContext.response, { errors: response.errors, data: response.data, extensions: response.extensions }),
                }).graphqlResponse);
            }
        });
    }
    initializeExtensionStack() {
        const extensions = this.config.extensions
            ? this.config.extensions.map(f => f())
            : [];
        if (this.config.tracing) {
            extensions.push(new apollo_tracing_1.TracingExtension());
        }
        let cacheControlExtension;
        if (this.config.cacheControl) {
            cacheControlExtension = new apollo_cache_control_1.CacheControlExtension(this.config.cacheControl);
            extensions.push(cacheControlExtension);
        }
        return new graphql_extensions_1.GraphQLExtensionStack(extensions);
    }
    initializeDataSources(requestContext) {
        if (this.config.dataSources) {
            const context = requestContext.context;
            const dataSources = this.config.dataSources();
            for (const dataSource of Object.values(dataSources)) {
                if (dataSource.initialize) {
                    dataSource.initialize({
                        context,
                        cache: requestContext.cache,
                    });
                }
            }
            if ('dataSources' in context) {
                throw new Error('Please use the dataSources config option instead of putting dataSources on the context yourself.');
            }
            context.dataSources = dataSources;
        }
    }
}
exports.GraphQLRequestPipeline = GraphQLRequestPipeline;
//# sourceMappingURL=requestPipeline.js.map