"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const graphql_1 = require("graphql");
const graphql_extensions_1 = require("graphql-extensions");
const apollo_cache_control_1 = require("apollo-cache-control");
const apollo_tracing_1 = require("apollo-tracing");
const apollo_server_errors_1 = require("apollo-server-errors");
const crypto_1 = require("crypto");
class InvalidGraphQLRequestError extends Error {
}
exports.InvalidGraphQLRequestError = InvalidGraphQLRequestError;
class GraphQLRequestProcessor {
    constructor(options) {
        this.options = options;
        this.context = options.context || {};
        this.initializeExtensions();
    }
    initializeExtensions() {
        const extensions = this.options.extensions
            ? this.options.extensions.map(f => f())
            : [];
        if (this.options.tracing) {
            extensions.push(new apollo_tracing_1.TracingExtension());
        }
        if (this.options.cacheControl === true) {
            extensions.push(new apollo_cache_control_1.CacheControlExtension());
        }
        else if (this.options.cacheControl) {
            extensions.push(new apollo_cache_control_1.CacheControlExtension(this.options.cacheControl));
        }
        this.extensionStack = new graphql_extensions_1.GraphQLExtensionStack(extensions);
        if (extensions.length > 0) {
            graphql_extensions_1.enableGraphQLExtensions(this.options.schema);
        }
        this.context._extensionStack = this.extensionStack;
    }
    processRequest(request) {
        return __awaiter(this, void 0, void 0, function* () {
            let { query, extensions } = request;
            let persistedQueryHit = false;
            let persistedQueryRegister = false;
            if (extensions && extensions.persistedQuery) {
                if (!this.options.persistedQueries ||
                    !this.options.persistedQueries.cache) {
                    throw new apollo_server_errors_1.PersistedQueryNotSupportedError();
                }
                else if (extensions.persistedQuery.version !== 1) {
                    throw new InvalidGraphQLRequestError('Unsupported persisted query version');
                }
                const sha = extensions.persistedQuery.sha256Hash;
                if (query === undefined) {
                    query =
                        (yield this.options.persistedQueries.cache.get(`apq:${sha}`)) ||
                            undefined;
                    if (query) {
                        persistedQueryHit = true;
                    }
                    else {
                        throw new apollo_server_errors_1.PersistedQueryNotFoundError();
                    }
                }
                else {
                    const hash = crypto_1.createHash('sha256');
                    const calculatedSha = hash.update(query).digest('hex');
                    if (sha !== calculatedSha) {
                        throw new InvalidGraphQLRequestError('provided sha does not match query');
                    }
                    persistedQueryRegister = true;
                    (() => __awaiter(this, void 0, void 0, function* () {
                        return (this.options.persistedQueries &&
                            this.options.persistedQueries.cache.set(`apq:${sha}`, query));
                    }))().catch(error => {
                        console.warn(error);
                    });
                }
            }
            if (!query) {
                throw new InvalidGraphQLRequestError('Must provide query string.');
            }
            const requestDidEnd = this.extensionStack.requestDidStart({
                request: request.httpRequest,
                queryString: request.query,
                operationName: request.operationName,
                variables: request.variables,
                extensions: request.extensions,
                persistedQueryHit,
                persistedQueryRegister,
            });
            try {
                let document;
                try {
                    document = this.parse(query);
                }
                catch (syntaxError) {
                    return this.willSendResponse({
                        errors: [
                            apollo_server_errors_1.fromGraphQLError(syntaxError, {
                                errorClass: apollo_server_errors_1.SyntaxError,
                            }),
                        ],
                    });
                }
                const validationErrors = this.validate(document);
                if (validationErrors.length > 0) {
                    return this.willSendResponse({
                        errors: validationErrors.map(validationError => apollo_server_errors_1.fromGraphQLError(validationError, {
                            errorClass: apollo_server_errors_1.ValidationError,
                        })),
                    });
                }
                const operation = graphql_1.getOperationAST(document, request.operationName);
                if (operation && this.willExecuteOperation) {
                    this.willExecuteOperation(operation);
                }
                let response;
                try {
                    response = (yield this.execute(document, request.operationName, request.variables));
                }
                catch (executionError) {
                    return this.willSendResponse({
                        errors: [apollo_server_errors_1.fromGraphQLError(executionError)],
                    });
                }
                const formattedExtensions = this.extensionStack.format();
                if (Object.keys(formattedExtensions).length > 0) {
                    response.extensions = formattedExtensions;
                }
                if (this.options.formatResponse) {
                    response = this.options.formatResponse(response, {
                        context: this.context,
                    });
                }
                return this.willSendResponse(response);
            }
            finally {
                requestDidEnd();
            }
        });
    }
    willSendResponse(response) {
        return this.extensionStack.willSendResponse({
            graphqlResponse: response,
        }).graphqlResponse;
    }
    parse(query) {
        const parsingDidEnd = this.extensionStack.parsingDidStart({
            queryString: query,
        });
        try {
            return graphql_1.parse(query);
        }
        finally {
            parsingDidEnd();
        }
    }
    validate(document) {
        let rules = graphql_1.specifiedRules;
        if (this.options.validationRules) {
            rules = rules.concat(this.options.validationRules);
        }
        const validationDidEnd = this.extensionStack.validationDidStart();
        try {
            return graphql_1.validate(this.options.schema, document, rules);
        }
        finally {
            validationDidEnd();
        }
    }
    execute(document, operationName, variables) {
        return __awaiter(this, void 0, void 0, function* () {
            const executionArgs = {
                schema: this.options.schema,
                document,
                rootValue: typeof this.options.rootValue === 'function'
                    ? this.options.rootValue(document)
                    : this.options.rootValue,
                contextValue: this.options.context,
                variableValues: variables,
                operationName,
                fieldResolver: this.options.fieldResolver,
            };
            const executionDidEnd = this.extensionStack.executionDidStart({
                executionArgs,
            });
            try {
                return graphql_1.execute(executionArgs);
            }
            finally {
                executionDidEnd();
            }
        });
    }
}
exports.GraphQLRequestProcessor = GraphQLRequestProcessor;
//# sourceMappingURL=requestProcessing.js.map